local Icon = LibStub("LibDBIcon-1.0")
local LDBobject = LibStub("LibDataBroker-1.1"):NewDataObject("BossList", {
	type = "launcher",
	icon = "Interface\\Buttons\\UI-GroupLoot-Dice-Up",
  OnTooltipShow = function(tooltip) 
    tooltip:AddLine("BossList")
    tooltip:AddLine("Click to open the overview of needs")
    tooltip:AddLine("Right click to open the settings") end,
  label = "BossList",
	OnClick = function() 
    if GetMouseButtonClicked() ==  "RightButton" then
      InterfaceOptionsFrame_OpenToCategory("BossList");
    else
      PlaySound("igMainMenuOptionCheckBoxOn")
      BossList.frame:Show()
    end end,
})

local AceConfig = LibStub("AceConfig-3.0")
local AceDB = LibStub("AceDB-3.0")
local DB

BossList.REVISION = 1

local frame = CreateFrame("Frame")
local DisplayEncounter_Orig
local currentInstanceID
local currentEncounterID
local dropDown

local desires = {"Need", "Minor", "Off-spec", "Not needed"}
local desireIcons = {"N", "M", "O", " "}
local desireColors = {{r=0, g=1, b=0}, {r=1, g=1, b=0}, {r=1, g=0.5, b=0}}

local defaults = {
  char = {
    minimap = {
      hide = false,
    }
  },
  global = {
    hideNotNeeded = true,
    blackList = {},
  }
}


function BossList_DEBUG_PrintEntriesIDs(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local msg = string.format("%d-%d-%d-%s-%d-%d", instance, difficulty, encounter, character, desire, lastUpdate)
            print(msg)
          end
        end
      end
    end
  end
end

function BossList_DEBUG_PrintEntries(table)
  print("Printing contents of table")
  local table = table or BossList_entries
  print(table)
  for instance, difficulties in pairs(table) do             -- Instances
    instanceName = EJ_GetInstanceInfo(instance) or "Unknown"
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        encounterName = EJ_GetEncounterInfo(encounter) or "Unknown"
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            lastUpdateTime = date("%d/%m/%y %H:%M:%S", lastUpdate) or "Unknown"
            local msg = string.format("%s-%d-%s-%s-%d-%s", instanceName, difficulty, encounterName, character, desire, lastUpdateTime)
            print(msg)
          end
        end
      end
    end
  end
end

local playerName = GetUnitName("player", false)

-- Extract the difficulty from the EncounterJournal link
local function GetDifficulty(link)
  local _, _, difficulty = link:find("%d+:%d+:(%d+)|")
  return tonumber(difficulty)
end

-- Might be used to catch changes between versions
local function CheckRevision()
  if DB.revision == nil or DB.revision < BossList.REVISION then
    --BossList_entries = {}
    --print("BossList: Wiped local database due to update")
    DB.revision = BossList.REVISION
    
  end
end

-- Return the entry if it exists, otherwise return nil
local function getEntry(table, instanceID, difficulty, encounterID, characterName)
  if table[instanceID] == nil or table[instanceID][difficulty] == nil or  -- Check if the encounter table exists
     table[instanceID][difficulty][encounterID] == nil then
    return nil
  end
  
  -- Check all desires for an entry
  for desire, characters in pairs(table[instanceID][difficulty][encounterID]) do
    if characters[characterName] ~= nil then
      return desire, characters[characterName]
    end
  end
  
  -- Nothing found
  return nil
end


-- Add entry into table
local function addEntry(table, instanceID, difficulty, encounterID, characterName, desirability, lastUpdate)
  if table[instanceID] == nil then table[instanceID] = {} end
  local instanceTable = table[instanceID]
  if instanceTable[difficulty] == nil then instanceTable[difficulty] = {} end
  local difficultyTable = instanceTable[difficulty]
  if difficultyTable[encounterID] == nil then difficultyTable[encounterID] = {} end
  local encounterTable = difficultyTable[encounterID]
  if encounterTable[desirability] == nil then encounterTable[desirability] = {} end
  local desireTable = encounterTable[desirability]
  
  desireTable[characterName] = lastUpdate or time()
end

-- Send a single entry to the guild
local function sendEntry(instance, difficulty, encounter, character, desirability, lastUpdate)
  local table = {}
  addEntry(table, instance, difficulty, encounter, character, desirability, lastUpdate)
  BossList.SendUpdate(table)
end

-- Compare all the entries in table to the local data. If newer or local is missing add it.
function BossList.CompareData(table)
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local localDesire, localLastUpdate = getEntry(BossList_entries, instance, difficulty, encounter, character)
            --print(string.format("Old entry: %d, New entry %d - %d %d %d %s\nNew desire: %d, old desire: %d", localLastUpdate, lastUpdate, instance, difficulty, encounter, character, desire, localDesire))
            if localDesire == nil or localLastUpdate <= lastUpdate then
              -- No local entry or received entry is newer, update local data.
              BossList.addEntry(instance, difficulty, encounter, character, desire, lastUpdate)
            end
          end
        end
      end
    end
  end
end

-- Compare all the entries in table to the local data. If newer or local is missing add it. Any local entries missing
-- in table are sent to the guild.
function BossList.MergeData(table)
  local updatesTable = {}
  for instance, difficulties in pairs(BossList_entries) do  -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            local recvDesire, recvLastUpdate = getEntry(table, instance, difficulty, encounter, character)
            if recvDesire == nil then
              -- Local entry exists but not contained in received table, add to table for broadcasting
              addEntry(updatesTable, instance, difficulty, encounter, character, desire, lastUpdate)
            else
              if recvLastUpdate < lastUpdate then
                -- Local entry is newer, add to table for broadcasting
                addEntry(updatesTable, instance, difficulty, encounter, character, desire, lastUpdate)
              else
                -- Received entry is newer, update local entry
                BossList.addEntry(instance, difficulty, encounter, character, recvDesire, recvLastUpdate)
              end
              table[instance][difficulty][encounter][recvDesire][character] = nil
            end
          end
        end
      end
    end
  end
  
  -- Send the updates to the guild
  if next(updatesTable) ~= nil then
    BossList.SendUpdate(updatesTable)
  end
  
  -- Check for any entries not in local table
  for instance, difficulties in pairs(table) do             -- Instances
    for difficulty, encounters in pairs(difficulties) do    -- Difficulties
      for encounter, desires in pairs(encounters) do        -- Encounters
        for desire, characters in pairs(desires) do         -- Desires
          for character, lastUpdate in pairs(characters) do -- Characters
            -- Insert in local table
            BossList.addEntry(instance, difficulty, encounter, character, desire, lastUpdate)
          end
        end
      end
    end
  end
end

-- Deletes entries older than 3 months
local function CheckTable()
  for instance, difficulties in pairs(BossList_entries) do -- Instances
    for difficulty, encounters in pairs(difficulties) do -- Difficulties
      for encounter, desires in pairs(encounters) do -- Encounters
        for desire, characters in pairs(desires) do -- Desires
          for character, lastUpdate in pairs(characters) do -- characters
            if time() - lastUpdate > 7776000 or BossList.IsBlackListed(character) then -- Older than 3 months
              
              BossList_entries[instance][difficulty][encounter][desire][character] = nil
            end
          end
          if next(characters) == nil then
            desires[desire] = nil
          end            
        end
        if next(desires) == nil then
          encounters[encounter] = nil
        end
      end
      if next(encounters) == nil then
        difficulties[encounter] = nil
      end
    end
    if next(difficulties) == nil then
      BossList_entries[instance] = nil
    end
  end
end

local function HideButtons()
  local _G = _G
  local counter = 1
  local BossEntry = _G["EncounterJournalBossButton"..counter]
  
  while BossEntry ~= nil do
    local bossButton = BossEntry.bossButton
    if bossButton ~= nil then
      bossButton:Hide()
    end

    counter = counter + 1
    BossEntry = _G["EncounterJournalBossButton"..counter]
  end
end

local function BossList_Menu_OnClick(self, desire, arg2, checked)  
  PlaySound("igMainMenuOptionCheckBoxOn")
  if not(IsInGuild()) then
    print("BossList: You are not in a guild, you can't change your choices")
    return
  end

  dropDown.button:SetText(desireIcons[desire])
  local color = desireColors[desire]
  if color ~= nil then
    dropDown.button:GetFontString():SetTextColor(color.r, color.g, color.b)
  end
  
  local encounterName, description, encounterID, rootSectionID, link = EJ_GetEncounterInfo(currentEncounterID)
  local instanceName = EJ_GetInstanceInfo()
  local difficulty = GetDifficulty(link)
  local characterName = playerName
  local lastUpdate = time()
  
  BossList.addEntry(currentInstanceID, difficulty, encounterID, characterName, desire, lastUpdate)
  sendEntry(currentInstanceID, difficulty, encounterID, characterName, desire, lastUpdate)
end

local function BossList_Menu(frame, level, menuList)
  local characterName = playerName
  local desire = 4
  if currentEncounterID ~= nil then
    local _, _, _, _, link = EJ_GetEncounterInfo(currentEncounterID)
    local entryDesire, _ = getEntry(BossList_entries, currentInstanceID, GetDifficulty(link), currentEncounterID, characterName)
    if entryDesire ~= nil then desire = entryDesire end
  end
  local info = UIDropDownMenu_CreateInfo()
  info.func = BossList_Menu_OnClick
  for idx, desireName in pairs(desires) do
    info.text, info.arg1, info.checked = desireName, idx, idx == desire
    UIDropDownMenu_AddButton(info)
  end
end

local function ToggleChoiceDropDown(button)
  PlaySound("igMainMenuOptionCheckBoxOn")
  currentEncounterID = button.encounterID
  dropDown.button = button
  ToggleDropDownMenu(1, nil, dropDown, "cursor", 3, -3)
end

local function DisplayInstance_Hook(instanceID, noButton)
  instanceID = instanceID or currentInstanceID
  --print("DisplayInstance" .. (instanceID or "nil"))
  if not EJ_InstanceIsRaid() then -- Only display if the instance is a raid
    return
  end
  
  local _G = _G
  local counter = 1
  local BossEntry = _G["EncounterJournalBossButton"..counter]
  local difficulty = GetDifficulty(BossEntry.link)
  while BossEntry ~= nil do
    local bossButton = BossEntry.bossButton
    if bossButton == nil then -- Create a button if it doesn't exist yet
      bossButton = CreateFrame("Button", "BossListbossButton", BossEntry, "EncounterDifficultyTemplate")
      
      bossButton:Raise()
      bossButton:SetScript("OnClick", ToggleChoiceDropDown)
      bossButton:SetPoint("RIGHT",BossEntry,-20,0)
      --bossButton:SetNormalTexture("Interface\\Buttons\\UI-GroupLoot-Dice-Up")
      BossEntry.bossButton = bossButton
    end
    
    local desire, _ = getEntry(BossList_entries, instanceID, difficulty, BossEntry.encounterID, playerName) or 4
    
    bossButton:SetText(desireIcons[desire])
    local color = desireColors[desire]
    if color ~= nil then
      bossButton:GetFontString():SetTextColor(color.r, color.g, color.b)
    end
    bossButton.encounterID = BossEntry.encounterID
    bossButton:Show()
    
    counter = counter + 1
    BossEntry = _G["EncounterJournalBossButton"..counter]
  end

  currentInstanceID = instanceID
end

local function Init()
  dropDown = CreateFrame("Frame", "BossListContextMenu", UIParent, "UIDropDownMenuTemplate")
  UIDropDownMenu_Initialize(dropDown, BossList_Menu, "MENU")
    
  hooksecurefunc("EncounterJournal_DisplayInstance", DisplayInstance_Hook);
  hooksecurefunc("EJ_SetDifficulty", function() DisplayInstance_Hook() end);
	hooksecurefunc(EncounterJournalNavBarHomeButton,"myclick", HideButtons);
end

function BossList.IsBlackListed(name)
  return tContains(DB.global.blackList, name)
end

local function EventHandler(self, event, ...)
  if event=="ADDON_LOADED" then
    local name = ...
    if name == "BossList" then   
      -- Create options
      DB = AceDB:New("BossList_DB", defaults)
      BossList.DB = DB
      
      if BossList_entries == nil then
        --print("BossList: First time loaded, initialized BossList_entries")
        BossList_entries = {}
      end
      CheckRevision()
      CheckTable() -- Check for old entries
      
      
      BossList.Synchronize()
      
      -- Create minimap button
      Icon:Register("BossList", LDBobject, DB.char.minimap)     
      
    
    elseif name == "Blizzard_EncounterJournal" then 
      DisplayEncounter_Orig = EncounterJournal_DisplayEncounter;
      
      Init()
      frame:UnregisterEvent("ADDON_LOADED")
    end
  end
end

-- Register chat prefixes
frame:RegisterEvent("ADDON_LOADED")
frame:SetScript("OnEvent", EventHandler)

optionsTable = {
  type = "group",
  args = {
    blackListHeader = {
      name = "Black list",
      order = 1,
      type = "header",
    },
    blackListDescription = {
      name = "Adding a player to the black list removes all his entries from BossList, and prevents new entries of this player from being added again. Beware, this also means you stop transmitting this player's entries to other players in the guild. As such, the black list should not be used as a filter.",
      fontSize = "medium",
      order = 2,
      type = "description",
    },    
    input = {
      name = "Add player",
      order = 3,
      width = "double",
      type = "input",
      pattern = "^[^%s%p%d%c]+$", -- No spaces, punctuation characters, digits or control characters allowed
      desc = "Enter a valid character name",
      set = function(info, string) 
              if not(tContains(DB.global.blackList, string)) then
                tinsert(DB.global.blackList, string)
                sort(DB.global.blackList)
                CheckTable()
              end
            end
    },
    blackList = {
      name = "Remove player",
      order = 4,
      type = "select",
      style = "dropdown",
      values = function() return DB.global.blackList end,
      set = function(info, val) 
              tremove(DB.global.blackList, val)
              sort(DB.global.blackList)
            end,
    },
    settingsHeader = {
      name = "Settings",
      order = 5,
      type = "header",
    },
    minimap = {
      name = "Hide minimap button",
      order = 6,
      desc = "Hides/shows the minimap button",
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                Icon:Hide("BossList") 
                DB.char.minimap.hide = true
              else 
                Icon:Show("BossList")
                DB.char.minimap.hide = false
              end end,
      get = function(info) return DB.char.minimap.hide end
    },
    notNeeded = {
      name = "Hide \"Not Needed\" entries",
      order = 7,
      desc = [[Hides/shows "Not Needed" in the BossList overview
      
Warning: decreases performance when opening/filtering the overview]],
      width = "double",
      type = "toggle",
      set = function(info,val)
              if val then 
                DB.global.hideNotNeeded = true
              else 
                DB.global.hideNotNeeded = false
              end 
              if BossList.frame.visible then
                BossList.frame:Show()
              end end,              
      get = function(info) return DB.global.hideNotNeeded end
    },
    
  }
}


LibStub("AceConfig-3.0"):RegisterOptionsTable("BossList", optionsTable);
LibStub("AceConfigDialog-3.0"):AddToBlizOptions("BossList", "BossList");

-- Register slash commands
SLASH_BOSSLIST1 = "/bosslist"
SLASH_BOSSLIST2 = "/bl"
SlashCmdList["BOSSLIST"] = function(msg)
  local arg1, arg2, arg3, arg4 = strsplit(" ", msg)
  if arg1 == "resync" then
    if arg2 ~= nil then
      RequestSync(arg2)
    else
      RequestUpdater()
    end
  else
    BossList.frame:Show()
  end
end